import numpy as np

def filter_nn(indices, dist, corr_thresh, signal2, signal_ref, centers_refplane, centers_plane2):
    match_index_array = np.ones((indices.shape[0],))*(-1) #index of cell in plane2 merged with corresponding row cell in refplane. -1 = unique
    for i, ix2 in enumerate(indices):
        if (np.corrcoef(signal2[int(ix2)][np.newaxis],signal_ref[i][np.newaxis])[0,1]>=corr_thresh) & (dist[i]<1):
            match_index_array[i] = ix2
            my_dist = np.sqrt((centers_refplane[i][0]-centers_plane2[int(ix2)][0])**2+(centers_refplane[i][1]-centers_plane2[int(ix2)][1])**2)
    return match_index_array

def nearest_neighbourds(centers_plane2, centers_refplane, radius):
    index_neighbours = np.ones(len(centers_refplane))-2
    distance_neighbours = np.ones(len(centers_refplane))*1000
    for idx, p_ref in enumerate(centers_refplane):
        current_min_dist = radius
        for idx2, p2 in enumerate(centers_plane2):
           dist = np.sqrt((p_ref[0]-p2[0])**2+(p_ref[1]-p2[1])**2)
           if dist<=current_min_dist:
               current_min_dist = dist
               index_neighbours[idx] = idx2
               distance_neighbours[idx] = dist          
    return distance_neighbours, index_neighbours

def merge_units(plane_name_2, plane_name_ref, dist_thresh, corr_thresh, CENTERS_dict, DFF_dict):
    centers_refplane = CENTERS_dict[plane_name_ref]
    centers_plane2 = CENTERS_dict[plane_name_2]
    dff_refplane = DFF_dict[plane_name_ref]
    dff_plane2 = DFF_dict[plane_name_2]
    dist, idx2 = nearest_neighbourds(centers_plane2, centers_refplane, radius = dist_thresh)
    #idx2 -= 1   # analysis.nearest_neighbour_assignmnet starts to number from 1, we adjust back to 0
    merged = filter_nn(idx2, dist, corr_thresh, dff_plane2, dff_refplane, centers_refplane, centers_plane2)
    return merged


import tifffile as tif
import numpy as np
import json
import os

import matplotlib.pyplot as plt
import warnings
from tqdm import tqdm
from matplotlib.backends.backend_pdf import PdfPages
import pandas as pd

import PyPDF2
from PyPDF2 import PdfReader
from reportlab.pdfgen import canvas
from reportlab.lib.pagesizes import letter
from reportlab.lib.colors import white, black

from pathlib import Path

from numan import utils, plots, analysis


def merge_pdfs(pdfs, filename):
    """
    Turns a bunch of separate figures (pdfs) into one pdf.
    """
    mergeFile = PyPDF2.PdfMerger()
    for pdf in pdfs:
        mergeFile.append(PyPDF2.PdfReader(pdf, 'rb'))
        os.remove(pdf)
    mergeFile.write(filename)


def place_cb(can, x, y, name):
    form = can.acroForm
    can.setFont("Courier", 12)
    can.drawCentredString(x + 20, y + 20, name)
    form.checkbox(name=name,
                  # tooltip = f"Field {name}",
                  x=x + 10,
                  y=y - 4,
                  # buttonStyle = 'check',
                  borderColor=black,
                  fillColor=white,
                  textColor=black,
                  forceBorder=True
                  )
    return can


def create_checkbox_pdf(pngs, cells_idx, btchs, filename):
    """
    Turns a bunch of separate figures (pngs) into one pdf with checkboxes.
    """
    # create an empty pdf
    can = canvas.Canvas(filename, pagesize=letter)
    # for each page:
    for ibtch, btch in enumerate(btchs):
        # refresh checkbox locations to be at the top,
        # numbers fitted to match the plots in the shape (5,1)
        X, Y, H = 10, 600, 122
        # add image to pdf and delete png file
        can.drawImage(pngs[ibtch], 0, 0, width=650,
                      preserveAspectRatio=True, mask='auto')
        # add checkboxes
        for cell_name in cells_idx[btch]:
            can = place_cb(can, x=X, y=Y, name=cell_name)
            Y = Y - H
        # finish page
        can.showPage()
        os.remove(pngs[ibtch])
    can.save()

def create_signal_pdf(project, experiment, spot_tag, group_tag, region_tag, saving_folder=None):

    print('Creating PDF with neurons signals...')
    if saving_folder is not None:
        os.makedirs(saving_folder, exist_ok=True) 
    report = Reports(Path(project.main_folder,"processed"), experiment)
    report.make_direct_signal_reports(spot_tag, group_tag,"number", region_tag,
                                labels= ["d0","d1","d2","d3","d4","d5"],
                                # types of plots:
                                plot_type = 'psh_b',
                                # which groups to write above the plots, also sorts but how many of these are significant
                                groups_to_specify=("anova_num_0","anova_num_1","anova_num_2","anova_num_3","anova_num_4","anova_num_5"),
                                # just for the pdf naming :
                                # this is to be able to distinguish the pdfs with the same plot type,
                                # but errors are different or raw traces on/off or front_to_tail
                                plot_type_tag = '',
                                t_points_before_stim = 3,
                                t_points_after_stim = 8,
                                # whether to plot the individual traces
                                plot_individual=False,
                                saving_folder = saving_folder)
 

def create_signal_pdf_new(reps_mean, reps_sem, coordinates, baseline_length, post_stim_length, resolution, output_pdf="calcium_signals_with_errorbars.pdf", neurons_per_page=5, neuron_filter=None, classify_neurons=False):
    """
    Creates a PDF with calcium signal (dF/F0) plots and error bars, with progress bar indication.
    
    Parameters:
    - reps_mean: np.array of shape (6, 12, n) containing the calcium signal means
    - reps_std: np.array of shape (6, 12, n) containing the calcium signal standard deviations
    - coordinates: list of lists of length n, where each element is a list [z, y, x] representing neuron positions
    - baseline_length: length of the baseline before stimulation
    - post_stim_length: length after stimulation
    - resolution: list of three values for the resolution [z_res, y_res, x_res]
    - output_pdf: name of the output PDF file
    - neurons_per_page: number of neurons per page (default: 5)
    - neuron_filter: boolean array of length n to filter which neurons to plot (default: None, plot all)
    - classify_neurons: If True, classify neurons and add colors and labels for preferred stimulus and excitatory/inhibitory.
    """
    num_neurons = reps_mean.shape[2]  # Total number of neurons

    # If no filter is provided, plot all neurons
    if neuron_filter is None:
        neuron_filter = np.ones(num_neurons, dtype=bool)

    # Select only the neurons that pass the filter
    selected_neurons = np.where(neuron_filter)[0]
    selected_coordinates = [coordinates[idx] for idx in selected_neurons]

    # Create x-ticks for the time points
    x_ticks = np.linspace(-baseline_length, post_stim_length, 12).astype(int)

    # Define background colors for classification with low alpha
    colors = ['red', 'gold', 'green', 'blue', 'purple', 'cyan']  ### change here colors, also the length of list if you've different n of stimulus types
    alpha = 0.2  # Set the transparency level

    # Initialize progress bar
    total_pages = len(selected_neurons) // neurons_per_page + (1 if len(selected_neurons) % neurons_per_page > 0 else 0)
    with tqdm(total=total_pages, desc="Generating PDF", unit="page") as pbar:
        # Create the PDF
        with PdfPages(output_pdf) as pdf:
            for i in range(0, len(selected_neurons), neurons_per_page):
                # Set up the figure for each page
                fig, axes = plt.subplots(neurons_per_page, 6, figsize=(15, neurons_per_page * 3)) ### change here 6 if you have different number of stimulus types (numerosities)
                fig.subplots_adjust(hspace=0.6, wspace=0.3)

                # Iterate over the neurons on the current page
                for j in range(neurons_per_page):
                    if i + j >= len(selected_neurons):
                        break  # Stop if we exceed the number of selected neurons

                    neuron_idx = selected_neurons[i + j]
                    coord = selected_coordinates[i + j]

                    # Convert coordinates using resolution
                    coord_converted = [int(coord[k] / resolution[k]) for k in range(3)]

                    # Neuron classification
                    if classify_neurons:
                        # Calculate the absolute average signal in the [baseline_length:baseline_length+3] window
                        averages = [np.abs(np.mean(reps_mean[stim_idx, baseline_length:baseline_length + 3, neuron_idx])) for stim_idx in range(6)] ### change here 6 if you have different number of stimulus types (numerosities)
                        preferred_stimulus_idx = np.argmax(averages)

                        # Check if the neuron is excitatory or inhibitory for the preferred stimulus
                        original_avg = np.mean(reps_mean[preferred_stimulus_idx, baseline_length:baseline_length + 3, neuron_idx])
                        neuron_type = 'excitatory' if original_avg > 0 else 'inhibitory'

                        # Add neuron type and classification to the title
                        neuron_title = f'Neuron {neuron_idx} (z={coord_converted[0]}, y={coord_converted[1]}, x={coord_converted[2]}) - {neuron_type} number {preferred_stimulus_idx} neuron'
                    else:
                        neuron_title = f'Neuron {neuron_idx} (z={coord_converted[0]}, y={coord_converted[1]}, x={coord_converted[2]})'

                    # Center the title above the third subplot (column 3)
                    axes[j, 2].set_title(neuron_title, fontsize=14, fontweight='bold', pad=20)  # Center title in third column

                    for stim_idx in range(6):  # Iterate over each stimulus type
                        ax = axes[j, stim_idx] if neurons_per_page > 1 else axes[stim_idx]

                        # If neuron classification is enabled, highlight the preferred stimulus with low alpha
                        if classify_neurons and stim_idx == preferred_stimulus_idx:
                            ax.set_facecolor(colors[preferred_stimulus_idx])
                            ax.patch.set_alpha(alpha)  # Set transparency level

                        # Plot signals with error bars
                        ax.errorbar(np.arange(12), reps_mean[stim_idx, :, neuron_idx],
                                    yerr=reps_sem[stim_idx, :, neuron_idx], fmt='-o')
                        ax.set_xlabel('time [s]' if j == neurons_per_page - 1 else '')  # X-label only on the last row
                        ax.set_ylabel('dF/F0' if stim_idx == 0 else '')  # Y-label only on the first plot of the row

                        # Set x-ticks to be integers from -baseline_length to +post_stim_length
                        ax.set_xticks(np.arange(12))  # Use original indices
                        ax.set_xticklabels(x_ticks)  # Set x-tick labels

                        # Add vertical dashed line for stimulus at baseline_length position
                        ax.axvline(x=baseline_length, color='red', linestyle='--', linewidth=1)

                # Save the page to the PDF
                pdf.savefig(fig)
                plt.close(fig)

                # Update the progress bar
                pbar.update(1)

    print(f"PDF saved as {output_pdf}")


class Reports:
    """
    For now it is simply a class of wrappers to make reports specifically for the 2vs3vs5 experiment.
    I hope it will become more clean and general as time goes on.
    """

    def __init__(self, project_folder, experiment):
        self.project = project_folder
        self.experiment = experiment

    @staticmethod
    def prepare_spot_info(spots, group_tag, n_traces, sort_by_sig=True,
                          groups_to_specify=("sig2v3", "sig2v5", "sig3v5", "sig2vB", "sig3vB", "sig5vB")):
        # some info on the cells to put into the title
        # cells idx in the original set of cells as they are in the spots
        cells_idx = spots.get_group_idx(spots.groups[group_tag])
        # cells coordinates
        cells_zyx = spots.get_group_centers(spots.groups[group_tag]).astype(np.int32)
        # list of groups that the cell belongs to ( from the list provided )
        cells_group = spots.get_group_info(groups_to_specify,
                                           group=spots.groups[group_tag])
        cells_group = np.array([group_name.replace("sig", "") for group_name in cells_group])
        # since signals are already cropped for the group, it's just np.arrange
        signal_idx = np.arange(n_traces)

        if sort_by_sig:
            # sort everything so that the cells with the most amount of significant stuff appear first
            sorted_zip = utils.sort_by_len0(zip(cells_group, cells_idx, cells_zyx, signal_idx))
            cells_group = np.array([el[0] for el in sorted_zip])
            cells_idx = np.array([el[1] for el in sorted_zip])
            cells_zyx = np.array([el[2] for el in sorted_zip])
            signal_idx = np.array([el[3] for el in sorted_zip])

        return cells_group, cells_idx, cells_zyx, signal_idx

    def make_signal_reports(self,
                            spot_tag, group_tag,
                            annotation_type,
                            dff_param,
                            labels=None,
                            plot_type="cycle",
                            plot_type_tag='',
                            forward_shift=0,
                            plot_individual=False,
                            groups_to_specify=("sig2v3", "sig2v5", "sig3v5", "sig2vB", "sig3vB", "sig5vB"),
                            tmp_folder=None,
                            pdf_filename=None,
                            checkbox=False):
        """
        Generates a pdf with the specified type of plots.

        :param checkbox:
        :type checkbox:
        :param groups_to_specify:
        :type groups_to_specify:
        :param labels:
        :type labels:
        :param annotation_type:
        :type annotation_type:
        :param dff_param: parameters for dff calculation (see Signals.as_dff).
            For example {"method":"sliding", "window":15} or {"method":"step", "step_size":9, "baseline_volumes": [0,1,2]}
        :type dff_param: dict
        :param spot_tag: what set of spots to use. Chooses the spots*.json based on this.
        :type spot_tag: str
        :param group_tag: what spots group to use.
        :type group_tag: str
        :param plot_type: what plot to output ["cycle","psh_0","psh_b"]
        :type plot_type: str
        :param plot_type_tag: just for the pdf naming : this is to be able to distinguish
                    the pdfs with the same plot type, but errors are different or raw traces on/off or front_to_tail...
        :type plot_type_tag: str
        :param forward_shift: forward_shift will shift the cycle by the set number of voxels
                    so when set to 3, there are 3 blank volumes at the begining and at the end ...
                    if set to 0, will have 6 leading blanks and will end right after the 5 dots (black bar)
        :type forward_shift: int
        :param time_points: only show certain timepoints from the signal, for example : only 2 dots.
                    IF time_points is 2d array, will overlap traces along axis = 1.
        :type time_points: numpy.array
        :param vlines: draw vertical lines, locations (in volumes) where to draw vertical lines
        :type vlines: list[int]
        :param signal_split: how to break the lines in the plot, this is relevant to the displayed x axis
        :type signal_split: numpy.array
        :param error_type: what error type to use ( "sem" for SEM or "prc" for 5th - 95th percentile )
        :type error_type: str
        :param noise_color: the color of the individual traces (if shown)
        :type noise_color: valid color definition
        :param plot_individual: wheather to plot the individual traces
        :type plot_individual: bool
        :param tmp_folder: will store batch images in this folder before merging pdf,
                    will be stored with reports if left None.
        :type tmp_folder: Union(str, Path)
        :param pdf_filename: the name of the pdf file to save, will be generated automatically if left None
        :type pdf_filename: str

        """
        if plot_type == "cycle":
            assert dff_param["method"] == "sliding", "Cycle plots only work with sliding dff"
        if plot_type == "psh_0" or plot_type == "psh_b":
            assert labels is not None, "Labels must not be None for psh-type plots"

        # fill out the defaults
        if tmp_folder is None:
            if checkbox:
                # where to temporary store images while the code is running
                tmp_folder = f"{self.project}/spots/reports/groupped/signals/"
            else:
                # where to temporary store images while the code is running
                tmp_folder = f"{self.project}/spots/reports/all_significant/signals/"

        if pdf_filename is None:
            if checkbox:
                # filename to save pdf with all the significant traces and checkboxes
                pdf_filename = f"{self.project}/spots/reports/groupped/signals/" \
                               f"CHOOSE_{plot_type}{plot_type_tag}_from_{spot_tag}_group_{group_tag}.pdf"
            else:
                # filename to save pdf with all the significant traces
                pdf_filename = f"{self.project}/spots/reports/all_significant/signals/" \
                               f"{plot_type}{plot_type_tag}_from_{spot_tag}_group_{group_tag}.pdf"

        spots = analysis.Spots.from_json(f"{self.project}/spots/signals/spots_{spot_tag}.json")

        # initialise the signal plotter with DFF signal
        print(f"Using the following parameters for signal DFF: {dff_param}")
        signals = spots.get_group_signals(spots.groups[group_tag]).as_dff(**dff_param)
        s_plotter = plots.SignalPlotter(signals, self.experiment, annotation_type)

        # prepare title info
        cells_group, cells_idx, cells_zyx, signal_idx = self.prepare_spot_info(spots, group_tag,
                                                                               s_plotter.n_traces,
                                                                               sort_by_sig=True,
                                                                               groups_to_specify=groups_to_specify)
        main_title = f"DFF signals, tscore image {spot_tag}, group {group_tag}"

        # choose traces per page
        if plot_type == "psh_0":
            if checkbox:
                raise AssertionError("Can not build checkboxes for plot type 'psh_0' ")
            tpp = 10  # traces per page
        else:
            tpp = 5

        # prepare the batches per page
        cells = np.arange(s_plotter.n_traces)
        btchs = [cells[s: s + tpp] for s in np.arange(np.ceil(s_plotter.n_traces / tpp).astype(int)) * tpp]

        plot_files = []

        for ibtch, btch in enumerate(btchs):

            if plot_type == "cycle":
                # titles for the current batch
                titles = [f"Cell {idx}, {group} XYZ : {zyx[2]},{zyx[1]},{zyx[0]} (voxel) "
                          for idx, group, zyx in zip(cells_idx[btch], cells_group[btch], cells_zyx[btch])]

                s_plotter.make_cycle_figure(signal_idx[btch],
                                            main_title,
                                            titles,
                                            forward_shift=forward_shift,
                                            # what grid to use to show the points
                                            figure_layout=[5, 1],
                                            # figure parameters
                                            figsize=(10, 12),
                                            dpi=60,
                                            # whether to plot the individual traces
                                            plot_individual=plot_individual)

            if plot_type == "psh_0":
                padding = [0]
                # titles for the current batch
                titles = [f"Cell {idx}, {group} \nXYZ : {zyx[2]},{zyx[1]},{zyx[0]} (voxel) "
                          for idx, group, zyx in zip(cells_idx[btch], cells_group[btch], cells_zyx[btch])]
                s_plotter.make_psh_figure(signal_idx[btch],
                                          labels, padding,
                                          main_title,
                                          titles,
                                          # what grid to use to show the points
                                          figure_layout=[5, 2],
                                          # figure parameters
                                          figsize=(10, 12),
                                          dpi=60,
                                          gridspec_kw={'hspace': 0.4, 'wspace': 0.3},
                                          # whether to plot the individual traces
                                          plot_individual=plot_individual, split=False)

            if plot_type == "psh_b":
                padding = [-3, -2, -1, 0, 1, 2, 3, 4, 5]
                # titles for the current batch
                titles = [f"Cell {idx}, {group} XYZ : {zyx[2]},{zyx[1]},{zyx[0]} (voxel) "
                          for idx, group, zyx in zip(cells_idx[btch], cells_group[btch], cells_zyx[btch])]
                s_plotter.make_psh_figure(signal_idx[btch],
                                          labels, padding,
                                          main_title,
                                          titles,
                                          # what grid to use to show the points
                                          figure_layout=[5, 1],
                                          # figure parameters
                                          figsize=(10, 12),
                                          dpi=60,
                                          # whether to plot the individual traces
                                          plot_individual=plot_individual)
            plt.xlabel('Volume in cycle')
            if checkbox:
                filename = f'{tmp_folder}signals_batch{ibtch}.png'
            else:
                filename = f'{tmp_folder}signals_batch{ibtch}.pdf'
            plt.savefig(filename)
            plt.close()
            plot_files.append(filename)

        if checkbox:
            cells_idx_txt = np.array([str(idx) for idx in cells_idx])
            create_checkbox_pdf(plot_files, cells_idx_txt, btchs, pdf_filename)
        else:
            merge_pdfs(plot_files, pdf_filename)

    def make_avg_intensity_reports(self,
                                   spot_tag, group_tag,
                                   annotation_type,
                                   dff_param,
                                   labels,
                                   number_cells=False,
                                   plot_type_tag='',
                                   pdf_filename=None):

        if pdf_filename is None:
            # filename to save pdf with all the significant traces
            pdf_filename = f"{self.project}/spots/reports/all_significant/signals/" \
                           f"avg_intensity_scatterplot{plot_type_tag}_from_{spot_tag}_group_{group_tag}.pdf"

        spots = analysis.Spots.from_json(f"{self.project}/spots/signals/spots_{spot_tag}.json")
        cells_idx = spots.get_group_idx(spots.groups[group_tag])

        # initialise the signal plotter with DFF signal
        print(f"Using the following parameters for signal DFF: {dff_param}")
        signals = spots.get_group_signals(spots.groups[group_tag]).as_dff(**dff_param)
        s_plotter = plots.SignalPlotter(signals, self.experiment, annotation_type)

        # prepare title info
        main_title = f"Average DFF per stimuli, pairwise comparison."

        # figure out figure layout
        if len(labels) == 3:
            figure_layout = [1, 3]
        elif len(labels) == 4:
            figure_layout = [2, 3]
        elif len(labels) == 5:
            figure_layout = [2, 5]
        else:
            raise Exception(f"Don't know how to plot when the number of labels is {len(labels)}")

        s_plotter.make_avg_act_scat_figure(labels, main_title,
                                           cell_numbers=cells_idx,
                                           # what grid to use to show the points
                                           figure_layout=figure_layout,
                                           # figure parameters
                                           figsize=(12, 10),
                                           dpi=160,
                                           number_cells=number_cells)

        plt.savefig(pdf_filename)
        plt.close()
        
    def make_direct_signal_reports(self,
                            spot_tag, group_tag, 
                            annotation_type, region,
                            labels=None,
                            plot_type="cycle",
                            plot_type_tag='',
                            t_points_before_stim = 3,
                            t_points_after_stim = 4,
                            forward_shift=0,
                            plot_individual=False,
                            groups_to_specify=("sig2v3", "sig2v5", "sig3v5", "sig2vB", "sig3vB", "sig5vB"),
                            tmp_folder=None,
                            pdf_filename=None,
                            checkbox=False,
                            saving_folder=None):
        """
        Generates a pdf with the specified type of plots.

        :param checkbox:
        :type checkbox:
        :param groups_to_specify:
        :type groups_to_specify:
        :param labels:
        :type labels:
        :param annotation_type:
        :type annotation_type:
        :param dff_param: parameters for dff calculation (see Signals.as_dff).
            For example {"method":"sliding", "window":15} or {"method":"step", "step_size":9, "baseline_volumes": [0,1,2]}
        :type dff_param: dict
        :param spot_tag: what set of spots to use. Chooses the spots*.json based on this.
        :type spot_tag: str
        :param group_tag: what spots group to use.
        :type group_tag: str
        :param plot_type: what plot to output ["cycle","psh_0","psh_b"]
        :type plot_type: str
        :param plot_type_tag: just for the pdf naming : this is to be able to distinguish
                    the pdfs with the same plot type, but errors are different or raw traces on/off or front_to_tail...
        :type plot_type_tag: str
        :param forward_shift: forward_shift will shift the cycle by the set number of voxels
                    so when set to 3, there are 3 blank volumes at the begining and at the end ...
                    if set to 0, will have 6 leading blanks and will end right after the 5 dots (black bar)
        :type forward_shift: int
        :param time_points: only show certain timepoints from the signal, for example : only 2 dots.
                    IF time_points is 2d array, will overlap traces along axis = 1.
        :type time_points: numpy.array
        :param vlines: draw vertical lines, locations (in volumes) where to draw vertical lines
        :type vlines: list[int]
        :param signal_split: how to break the lines in the plot, this is relevant to the displayed x axis
        :type signal_split: numpy.array
        :param error_type: what error type to use ( "sem" for SEM or "prc" for 5th - 95th percentile )
        :type error_type: str
        :param noise_color: the color of the individual traces (if shown)
        :type noise_color: valid color definition
        :param plot_individual: wheather to plot the individual traces
        :type plot_individual: bool
        :param tmp_folder: will store batch images in this folder before merging pdf,
                    will be stored with reports if left None.
        :type tmp_folder: Union(str, Path)
        :param pdf_filename: the name of the pdf file to save, will be generated automatically if left None
        :type pdf_filename: str

        """

        # fill out the defaults
        if tmp_folder is None:
            if checkbox:
                # where to temporary store images while the code is running
                tmp_folder = f"{self.project}/spots/reports/"
            else:
                # where to temporary store images while the code is running
                tmp_folder = f"{self.project}/spots/reports/"

        if pdf_filename is None:
            if checkbox:
                # filename to save pdf with all the significant traces and checkboxes
                if saving_folder is None:
                    pdf_filename = f"{self.project}/spots/reports/" \
                                f"CHOOSE_{plot_type}{plot_type_tag}_from_{spot_tag}_group_{group_tag}_{region}.pdf"
                else:
                    pdf_filename = f"{self.project}/spots/reports/{saving_folder}/" \
                                f"CHOOSE_{plot_type}{plot_type_tag}_from_{spot_tag}_group_{group_tag}_{region}.pdf"

            else:
                # filename to save pdf with all the significant traces
                if saving_folder:
                    pdf_filename = f"{self.project}/spots/reports/{saving_folder}/" \
                               f"{plot_type}{plot_type_tag}_from_{spot_tag}_group_{group_tag}_{region}.pdf"
                else:
                    pdf_filename = f"{self.project}/spots/reports/" \
                               f"{plot_type}{plot_type_tag}_from_{spot_tag}_group_{group_tag}_{region}.pdf"

        spots = analysis.Spots.from_json(f"{self.project}/spots/signals/spots_{spot_tag}.json")

        # initialise the signal plotter with DFF signal
        signals = spots.get_group_signals(spots.groups[group_tag])
        s_plotter = plots.SignalPlotter(signals, self.experiment, annotation_type)

        # prepare title info
        cells_group, cells_idx, cells_zyx, signal_idx = self.prepare_spot_info(spots, group_tag,
                                                                               s_plotter.n_traces,
                                                                               sort_by_sig=True,
                                                                               groups_to_specify=groups_to_specify)
        main_title = f"DFF signals, tscore image {spot_tag}, group {group_tag}"

        # choose traces per page
        if plot_type == "psh_0":
            if checkbox:
                raise AssertionError("Can not build checkboxes for plot type 'psh_0' ")
            tpp = 10  # traces per page
        else:
            tpp = 5

        # prepare the batches per page
        cells = np.arange(s_plotter.n_traces)
        btchs = [cells[s: s + tpp] for s in np.arange(np.ceil(s_plotter.n_traces / tpp).astype(int)) * tpp]

        plot_files = []

        for ibtch, btch in enumerate(btchs):

            if plot_type == "cycle":
                # titles for the current batch
                titles = [f"Cell {idx}, {group} XYZ : {zyx[2]},{zyx[1]},{zyx[0]} (voxel) "
                          for idx, group, zyx in zip(cells_idx[btch], cells_group[btch], cells_zyx[btch])]

                s_plotter.make_cycle_figure(signal_idx[btch],
                                            main_title,
                                            titles,
                                            forward_shift=forward_shift,
                                            # what grid to use to show the points
                                            figure_layout=[5, 1],
                                            # figure parameters
                                            figsize=(10, 12),
                                            dpi=60,
                                            # whether to plot the individual traces
                                            plot_individual=plot_individual)

            if plot_type == "psh_0":
                padding = [0]
                # titles for the current batch
                titles = [f"Cell {idx}, {group} \nXYZ : {zyx[2]},{zyx[1]},{zyx[0]} (voxel) "
                          for idx, group, zyx in zip(cells_idx[btch], cells_group[btch], cells_zyx[btch])]
                s_plotter.make_psh_figure(signal_idx[btch],
                                          labels, padding,
                                          main_title,
                                          titles,
                                          # what grid to use to show the points
                                          figure_layout=[5, 2],
                                          # figure parameters
                                          figsize=(10, 12),
                                          dpi=60,
                                          gridspec_kw={'hspace': 0.4, 'wspace': 0.3},
                                          # whether to plot the individual traces
                                          plot_individual=plot_individual, split=False)

            if plot_type == "psh_b":
                padding = list(range(-t_points_before_stim, t_points_after_stim + 1))
                # titles for the current batch
                titles = [f"Cell {idx}, {group} XYZ : {zyx[2]},{zyx[1]},{zyx[0]} (voxel) "
                          for idx, group, zyx in zip(cells_idx[btch], cells_group[btch], cells_zyx[btch])]
                s_plotter.make_psh_figure(signal_idx[btch],
                                          labels, padding,
                                          main_title,
                                          titles,
                                          # what grid to use to show the points
                                          figure_layout=[5, 1],
                                          # figure parameters
                                          figsize=(10, 12),
                                          dpi=60,
                                          # whether to plot the individual traces
                                          plot_individual=plot_individual)
            plt.xlabel('Volume in cycle')
            if checkbox:
                filename = f'{tmp_folder}signals_batch{ibtch}.png'
            else:
                filename = f'{tmp_folder}signals_batch{ibtch}.pdf'
            plt.savefig(filename)
            plt.close()
            plot_files.append(filename)

        if checkbox:
            cells_idx_txt = np.array([str(idx) for idx in cells_idx])
            create_checkbox_pdf(plot_files, cells_idx_txt, btchs, pdf_filename)
        else:
            merge_pdfs(plot_files, pdf_filename)

    def make_covariate_reports(self,
                               spot_tag, group_tag,
                               annotation_type,
                               dff_param,
                               conditions=None,
                               plot_type="cycle",
                               plot_type_tag='',
                               forward_shift=0,
                               plot_individual=False,
                               groups_to_specify=("sig2v3", "sig2v5", "sig3v5", "sig2vB", "sig3vB", "sig5vB"),
                               tmp_folder=None,
                               pdf_filename=None):
        """
        Generates a pdf with the specified type of plots.

        :param checkbox:
        :type checkbox:
        :param groups_to_specify:
        :type groups_to_specify:
        :param conditions:
        :type conditions:
        :param annotation_type:
        :type annotation_type:
        :param spot_tag: what set of spots to use. Chooses the spots*.json based on this.
        :type spot_tag: str
        :param group_tag: what spots group to use.
        :type group_tag: str
        :param plot_type: what plot to output ["cycle","psh_0","psh_b"]
        :type plot_type: str
        :param plot_type_tag: just for the pdf naming : this is to be able to distinguish
                    the pdfs with the same plot type, but errors are different or raw traces on/off or front_to_tail...
        :type plot_type_tag: str
        :param forward_shift: forward_shift will shift the cycle by the set number of voxels
                    so when set to 3, there are 3 blank volumes at the begining and at the end ...
                    if set to 0, will have 6 leading blanks and will end right after the 5 dots (black bar)
        :type forward_shift: int
        :param time_points: only show certain timepoints from the signal, for example : only 2 dots.
                    IF time_points is 2d array, will overlap traces along axis = 1.
        :type time_points: numpy.array
        :param vlines: draw vertical lines, locations (in volumes) where to draw vertical lines
        :type vlines: list[int]
        :param signal_split: how to break the lines in the plot, this is relevant to the displayed x axis
        :type signal_split: numpy.array
        :param error_type: what error type to use ( "sem" for SEM or "prc" for 5th - 95th percentile )
        :type error_type: str
        :param noise_color: the color of the individual traces (if shown)
        :type noise_color: valid color definition
        :param plot_individual: wheather to plot the individual traces
        :type plot_individual: bool
        :param tmp_folder: will store batch images in this folder before merging pdf,
                    will be stored with reports if left None.
        :type tmp_folder: Union(str, Path)
        :param pdf_filename: the name of the pdf file to save, will be generated automatically if left None
        :type pdf_filename: str

        """
        if plot_type == "cycle":
            assert dff_param["method"] == "sliding", "Cycle plots only work with sliding dff"

        assert conditions is not None, "Conditions must not be None for psh-type plots"

        # fill out the defaults
        if tmp_folder is None:
            # where to temporary store images while the code is running
            tmp_folder = f"{self.project}/spots/reports/covariates/signals/"

        if pdf_filename is None:
            # filename to save pdf with all the significant traces
            pdf_filename = f"{self.project}/spots/reports/covariates/signals/" \
                           f"{plot_type}{plot_type_tag}_from_{spot_tag}_group_{group_tag}.pdf"

        spots = analysis.Spots.from_json(f"{self.project}/spots/signals/spots_{spot_tag}.json")

        # initialise the signal plotter with DFF signal
        print(f"Using the following parameters for signal DFF: {dff_param}")
        signals = spots.get_group_signals(spots.groups[group_tag]).as_dff(**dff_param)

        # choose traces per page
        if plot_type == "psh_0":
            tpp = 10  # traces per page
            s_plotter = plots.SignalPlotter(signals, self.experiment, annotation_type,
                                      c_mean_color='w', c_noise_color='-m', c_edge_color='w')
        else:
            tpp = 5
            s_plotter = plots.SignalPlotter(signals, self.experiment, annotation_type,
                                      c_mean_color='k', c_noise_color='-m', c_edge_color='w')

        # prepare title info
        cells_group, cells_idx, cells_zyx, signal_idx = self.prepare_spot_info(spots, group_tag,
                                                                               s_plotter.n_traces,
                                                                               sort_by_sig=True,
                                                                               groups_to_specify=groups_to_specify)
        main_title = f"DFF signals, tscore image {spot_tag}, group {group_tag}"

        # prepare the batches per page
        cells = np.arange(s_plotter.n_traces)
        btchs = [cells[s: s + tpp] for s in np.arange(np.ceil(s_plotter.n_traces / tpp).astype(int)) * tpp]

        plot_files = []

        for ibtch, btch in enumerate(btchs):

            if plot_type == "psh_0":
                padding = [0]
                # titles for the current batch
                titles = [f"Cell {idx}, {group} \nXYZ : {zyx[2]},{zyx[1]},{zyx[0]} (voxel) "
                          for idx, group, zyx in zip(cells_idx[btch], cells_group[btch], cells_zyx[btch])]
                s_plotter.make_covariate_psh_figure(signal_idx[btch],
                                                    conditions, padding,
                                                    main_title,
                                                    titles,
                                                    # what grid to use to show the points
                                                    figure_layout=[5, 2],
                                                    # figure parameters
                                                    figsize=(10, 12),
                                                    dpi=60,
                                                    gridspec_kw={'hspace': 0.4, 'wspace': 0.3},
                                                    # whether to plot the individual traces
                                                    plot_individual=plot_individual, split=False)

            if plot_type == "psh_b":
                padding = [-3, -2, -1, 0, 1, 2, 3, 4, 5]
                # titles for the current batch
                titles = [f"Cell {idx}, {group} XYZ : {zyx[2]},{zyx[1]},{zyx[0]} (voxel) "
                          for idx, group, zyx in zip(cells_idx[btch], cells_group[btch], cells_zyx[btch])]
                s_plotter.make_covariate_psh_figure(signal_idx[btch],
                                                    conditions, padding,
                                                    main_title,
                                                    titles,
                                                    # what grid to use to show the points
                                                    figure_layout=[5, 1],
                                                    # figure parameters
                                                    figsize=(10, 12),
                                                    dpi=60,
                                                    # whether to plot the individual traces
                                                    plot_individual=plot_individual)
            plt.xlabel('Volume in cycle')
            filename = f'{tmp_folder}signals_batch{ibtch}.pdf'
            plt.savefig(filename)
            plt.close()
            plot_files.append(filename)

        merge_pdfs(plot_files, pdf_filename)


class CellMasks:
    def __init__(self, project, experiment):
        self.project = project
        self.experiment = experiment
        self.filenames, self.new_groups, self.spot_tags = self.get_new_groups()

    def get_new_groups(self, verbose=True):
        group_filenames = list(Path(self.project.main_folder,
                                    'processed/spots/reports/groupped/signals/').glob('[*'))
        new_groups = [str(name).split('[')[1].split(']')[0] for name in group_filenames]
        spot_tags = [str(name).split('from_')[1].split('_group')[0] for name in group_filenames]
        if verbose:
            print(pd.DataFrame({"new group": new_groups, "from spots": spot_tags}))
        return group_filenames, new_groups, spot_tags

    @staticmethod
    def get_checked_boxes(filename):
        """get checked boxes from pdf"""

        f = open(filename, "rb")
        group_pdf = PyPDF2.PdfReader(f)
        fields = group_pdf.get_fields()
        spots_in_group = np.array([spot for spot in fields if
                                   fields[spot]['/V'] == '/Yes']).astype(int)
        f.close()
        return spots_in_group

    def add_groups_from_pdf(self, rewrite=False):
        spots = analysis.Spots.from_json(f"spots/signals/spots_{self.spot_tags[0]}.json")
        current_spots = self.spot_tags[0]
        for filename, group_tag, spot_tag in zip(self.filenames, self.new_groups, self.spot_tags):
            if spot_tag != current_spots:
                spots.to_json(f"spots/signals/spots_{current_spots}.json")
                # prepare the new group
                current_spots = spot_tag
                spots = Spots.from_json(f"spots/signals/spots_{spot_tag}.json")

            # get checked boxes
            spots_in_group = self.get_checked_boxes(filename)
            is_in_group = np.zeros((spots.num_spots,)).astype(bool)
            is_in_group[spots_in_group] = True

            # add the new groups to spots
            spots.add_groups({group_tag: is_in_group}, rewrite=rewrite)
            print(f"Added group {group_tag} to spots_{spot_tag}.json")
        # save the last chunk
        spots.to_json(f"spots/signals/spots_{current_spots}.json")

    def write_masks(self, resolution):
        # get the size of one volume
        T, Z, Y, X = self.experiment.load_volumes([0]).shape
        print(f"Image shape : {T}, {Z}, {Y}, {X}")

        # resolution in ZYX order in um
        image_dir = Path(self.project.main_folder, "processed", "spots", "reports", "groupped", "images")

        spots = analysis.Spots.from_json(f"spots/signals/spots_{self.spot_tags[0]}.json")
        current_spots = self.spot_tags[0]
        for filename, group_tag, spot_tag in zip(self.filenames, self.new_groups, self.spot_tags):
            if spot_tag != current_spots:
                current_spots = spot_tag
                spots = analysis.Spots.from_json(f"spots/signals/spots_{spot_tag}.json")

            mask = spots.get_group_mask(spots.groups[group_tag], (Z, Y, X))
            tif.imwrite(f'{image_dir}/mask_from_{spot_tag}_group_{group_tag}.tif',
                        mask.astype(np.uint16), shape=(Z, Y, X),
                        metadata={'spacing': resolution[0], 'unit': 'um', 'axes': 'ZYX'},
                        resolution=(1 / resolution[1], 1 / resolution[2]), imagej=True)
            filename.rename(Path(filename.parent, f"done_{filename.name}"))
            print(f"Created mask for spots from {spot_tag} tscore image, for group {group_tag}")



import numpy as np
import numpy.typing as npt
import pandas as pd
from statistics import mean, stdev, sqrt
import scipy.stats as stats
from tqdm.notebook import tqdm
import tifffile as tif
from sklearn.utils import shuffle
import matplotlib.pyplot as plt
import os

from numan_plus import compute_tunings

# Vectorized implementation of two-way ANOVA
def anova_two_way(A, B, Y):
    num_cells = Y.shape[1]
    
    A_levels = np.unique(A); a = len(A_levels)
    B_levels = np.unique(B); b = len(B_levels)
    Y4D = np.array([[Y[(A==i)&(B==j)] for j in B_levels] for i in A_levels])
    
    r = Y4D.shape[2]

    Y = Y4D.reshape((-1, Y.shape[1]))
    
    # only test cells (units) that are active (gave a nonzero response to at least one stimulus) to avoid division by zero errors
    active_cells = np.where(np.abs(Y).max(axis=0)>0)[0]
    Y4D = Y4D[:,:,:,active_cells]
    Y = Y[:, active_cells]
    
    N = Y.shape[0]
    
    Y_mean = Y.mean(axis=0)
    Y_mean_A = Y4D.mean(axis=1).mean(axis=1)
    Y_mean_B = Y4D.mean(axis=0).mean(axis=1)
    Y_mean_AB = Y4D.mean(axis=2)

    
    SSA = r*b*np.sum((Y_mean_A - Y_mean)**2, axis=0)
    SSB = r*a*np.sum((Y_mean_B - Y_mean)**2, axis=0)
    SSAB = r*((Y_mean_AB - Y_mean_A[:,None] - Y_mean_B[None,:] + Y_mean)**2).sum(axis=0).sum(axis=0)
    SSE = ((Y4D-Y_mean_AB[:,:,None])**2).sum(axis=0).sum(axis=0).sum(axis=0)
    SST = ((Y-Y_mean)**2).sum(axis=0)

    DFA = a - 1; DFB = b - 1; DFAB = DFA*DFB
    DFE = (N-a*b); DFT = N-1
    
    MSA = SSA / DFA
    MSB = SSB / DFB
    MSAB = SSAB / DFAB
    MSE = SSE / DFE
    
    FA = MSA / MSE
    FB = MSB / MSE
    FAB = MSAB / MSE
    
    pA = np.nan*np.zeros(num_cells)
    pB = np.nan*np.zeros(num_cells)
    pAB = np.nan*np.zeros(num_cells)
    
    pA[active_cells] = stats.f.sf(FA, DFA, DFE)
    pB[active_cells] = stats.f.sf(FB, DFB, DFE)
    pAB[active_cells] = stats.f.sf(FAB, DFAB, DFE)
    
    return pA, pB, pAB, FA, FB, FAB


def anova_two_way_permutations(A, B, Y, num_perm, show_progress=True):
    a, b, c, FA0, FB0, FAB0 = anova_two_way(A, B, Y)
    num_cells = Y.shape[1]

    A_levels = np.unique(A); a = len(A_levels)
    B_levels = np.unique(B); b = len(B_levels)
    Y4D = np.array([[Y[(A==i)&(B==j)] for j in B_levels] for i in A_levels])
    
    r = Y4D.shape[2]

    Y = Y4D.reshape((-1, Y.shape[1]))
    
    active_cells = np.where(np.abs(Y).max(axis=0)>0)[0]
    Y4D = Y4D[:,:,:,active_cells]
    Y = Y[:, active_cells]

    FA0 = np.expand_dims(FA0, axis=1)
    FB0 = np.expand_dims(FB0, axis=1)
    FAB0 = np.expand_dims(FAB0, axis=1)
    nperm = num_perm
    FA = np.nan*np.zeros((active_cells.shape[0], nperm))
    FB = np.nan*np.zeros((active_cells.shape[0], nperm))
    FAB = np.nan*np.zeros((active_cells.shape[0], nperm))

    perm_iter = range(nperm)
    if show_progress:
        perm_iter = tqdm(perm_iter, desc='Permutations')
    
    for i in perm_iter:
        np.random.shuffle(Y)
        a, b, c, FA[:,i], FB[:,i], FAB[:,i] = anova_two_way(A, B, Y)

    pA = np.nan*np.zeros(num_cells)
    pB = np.nan*np.zeros(num_cells)
    pAB = np.nan*np.zeros(num_cells)

    pA[active_cells] = np.sum(np.greater_equal(FA, FA0), axis=1) / nperm
    pB[active_cells] = np.sum(np.greater_equal(FB, FB0), axis=1) / nperm
    pAB[active_cells] = np.sum(np.greater_equal(FAB, FAB0), axis=1) / nperm

    return pA, pB, pAB

def group_logical_and(group1, group2):
    """
    Performs element-wise logical and
    on two boolean arrays expects 1D or 2D (with second dimention of size 1) boolean arrays.
    Returns a 1D boolean array.
    """
    if len(group1.shape) >1: 
        assert len(group1.shape)==2, "expects less than 2 dimension but got more than 2 dimension"
        group1 = np.squeeze(group1)

    if len(group2.shape) >1: 
        assert len(group2.shape) ==2, "expects less than 2 dimension but got more than 2 dimension"
        group2 = np.squeeze(group2)

    assert group1.shape == group2.shape , "Dimentions of things for logical and must match"

    return np.logical_and(group1,group2)


def get_peristim(experiment, timepoints_to_use:list, stimulus_type:tuple, signals:npt.NDArray)->npt.NDArray: # add signals to argument
    """
    Makes a 3d array (timepoints, peristimulus cycle, cells) 
    from experiment_truncated_drift_corrected.db and json signal
    Args:
        timepoints_to_use: list, grabs indexed around the stimulus, +/- is before and after
        stimulus_type: tuple, get the stimulus you want from experiment_truncated_drift_corrected
            All of a certain number or shape
        signals: numpy array, the signal trace
    Returns:
        3D Numpy array of (timepoints, peristimulus cycle, cells)
    """
    idx = experiment.choose_volumes(stimulus_type)
    #make empty list
    idx_block = []

    #makes a list of indexes to get signal later
    idx_block = [i + j for i in idx for j in timepoints_to_use]

    #make a list of signals for each stimulus
    #order will be (timepoints, peristimulus cycle, cells)
    number_per_peristim = len(timepoints_to_use)
    number_of_peristim_cycle = int(len(idx_block)/number_per_peristim)
    #signal is a numpy array and downstream stuff are too so dn blocks will be too
    block_signal = np.empty((len(idx_block), signals.shape[1]))

    #print(f'Note: you are exctracting signal from {signals.shape[1]} neurons')
    # for foo in np.arange(signals.shape[1]):
    for neur in np.arange(signals.shape[1]):
        for t_point in np.arange(len(idx_block)):
            block_signal[t_point, neur] = signals[idx_block[t_point], neur]
    # print(block_signal.shape)

    #reshape so we can easily find the mean later
    #block_signal_reshaped = block_signal.reshape(number_per_peristim, number_of_peristim_cycle, signals.shape[1])
    block_signal_reshaped = block_signal.reshape(number_of_peristim_cycle, number_per_peristim, signals.shape[1])
    return block_signal_reshaped

def ANOVA_preprocess (experiment, stim_signal_exact, brain_region_tag, stim_volumes, signals):
    """
    Crate following variables for ANOVA calculation:
    Hf: matrix of responses (cells X trials); Q: array of type of stimulus (trials,); C: array of type of control condition (trials,)
    The responses are evaluated as avg across 3 time points from stimulus onset
    """

    ## take a single avg value as reference response to stimulus (avg across 3 vols from stimulation)
    stim_signal_prov = np.zeros((stim_signal_exact.shape[0],stim_signal_exact.shape[1],3))
    stim_signal_prov[:,:,0] = stim_signal_exact
    for i in [1,2]: # add additional volume after stimulus to calculate final avg signal
        stim_add_volumes = [x+i for x in stim_volumes]
        stim_add_signal = signals[:,stim_add_volumes]
        stim_signal_prov[:,:,i] = stim_add_signal
    ## create final matrix Hf for anova (cells X trials)
    stim_signal = stim_signal_prov.mean(axis=2)
    print('You are passing to the ANOVA, (cells,trials): ' + str(stim_signal.shape)+'\n')
    annotation_dict2= {f"cell_{ic}": stim_signal[ic] for ic in np.arange(len(signals))}
    annotation_dict=experiment.get_volume_annotations(stim_volumes)
    annotation_dict.update(annotation_dict2)
    annotation_df=pd.DataFrame(annotation_dict)
    Hf = np.array(annotation_df.iloc[:, 4:annotation_df.shape[1]])
    #print('Final dataset, shape -> (trials,cells): ' + str(Hf.shape))

    #calculate control trials label array C for anova (trials,): in our case array of 0,1,2,3,4,5 corresponding to the 6 combinations of getical control conditions (shape*spread)
    C_pd = pd.factorize((annotation_df['shape']+ annotation_df['spread']), sort=True)
    C = np.array(C_pd[0])
    #print('Control conditions label array, shape -> (trials,): ' + str(C.shape))

    #calculate stimulus trials label array Q for anova (trials,): in our case array of 0,1,2,3,4 corresponding to the 5 numerosity
    Q_pd = pd.factorize(annotation_df['number'], sort=True)
    Q = np.array(Q_pd[0])
    #print('Stimulus label array, shape -> (trials,): ' + str(Q.shape))

    return Hf, C, Q


def compute_anova_neurons(Q, C, Hf, alpha_level, n_permutations, filtered_idx):#, brain_region_tag, save_df):
    print('\nRunning permutation ANOVA on real dataset:')

    # Find numorosity selective units (anova_cells) using a two-way ANOVA with permutations (permute data and check F distribution for p-value)
    pN, pC, pNC = anova_two_way_permutations(Q, C, Hf, n_permutations, show_progress=True)
    anova_cells = np.where((pN<alpha_level) & (pNC>alpha_level) & (pC>alpha_level))[0]
    R = Hf[:,anova_cells]
    #save_df['Total segmented'] = [Hf.shape[1]]
    #save_df['Anova selective'] = [R.shape[1]]

    chance_lev = Hf.shape[1]*alpha_level/6
    #save_df['chance n cells per group'] = [chance_lev]
    #print('Chance number of cells for group: '+str(chance_lev))
    print('Number of anova cells = %i (%0.2f%%)'%(len(anova_cells), 100*len(anova_cells)/Hf.shape[1]))

    ##Creating dictionary for plotting anova cells: 'cell_ID'=preferred numerosity
    pref_num, excitatory_or_inhibitory = compute_tunings.preferred_numerosity(Q, R)
    number_cells_dic = {'anova_cells': filtered_idx[anova_cells], 'pref_num': pref_num, 'excitatory_or_inhibitory': excitatory_or_inhibitory}
    anova_df = pd.DataFrame.from_dict(data=number_cells_dic)
    #for n in range(6):
    #    save_df[f'Preferring_{n}'] = [sum(pref_num==n)]
    #
    #os.makedirs('./caiman_final_datasets', exist_ok=True) 
    #anova_df.to_csv(f'./caiman_final_datasets/numerosityCells_{brain_region_tag}.csv')
    #print('\033[1m\nYour number units are calculated.\033[0m\nYou can find them in ./processed/caiman_final_dataset')

    return anova_cells, R, chance_lev, anova_df, pref_num, excitatory_or_inhibitory#, save_df

def compute_shuffled_anova_neurons(Q, C, Hf, alpha_level, n_permutations, show_progress=False):
    if show_progress:
        print('\nRunning permutation ANOVA on shuffled dataset:')

    # Shuffle the datasets
    Q_S = shuffle(Q, random_state=0)
    C_S = shuffle(C, random_state=0)

    # Perform two-way ANOVA with permutations
    pN_s, pC_s, pNC_s = anova_two_way_permutations(Q_S, C_S, Hf, n_permutations, show_progress)
    
    # Identify selective units based on p-values
    anova_cells_shuffled = np.where((pN_s < alpha_level) & (pNC_s > alpha_level) & (pC_s > alpha_level))[0]
    R_S = Hf[:, anova_cells_shuffled]

    if show_progress:
        print('Number of ANOVA cells on random shuffled data = %i (%0.2f%%)' % 
              (len(anova_cells_shuffled), 100 * len(anova_cells_shuffled) / Hf.shape[1]))

    # Compute preferred numerosity for the selected neurons
    pref_num_shuffled, excitatory_or_inhibitory_shuffled = compute_tunings.preferred_numerosity(Q_S, R_S)

    return anova_cells_shuffled, Q_S, C_S, R_S, pref_num_shuffled, excitatory_or_inhibitory_shuffled

def run_multiple_shuffles(Q, C, Hf, alpha_level, n_permutations, n_reps=1000, show_inner_progress=False):
    n_numerosities = 6  # Define the number of preferred numerosity levels
    n_neurons_shuffled_excitatory = np.zeros((n_numerosities, n_reps))
    n_neurons_shuffled_inhibitory = np.zeros((n_numerosities, n_reps))
    n_neurons_shuffled_total = np.zeros((n_numerosities, n_reps))

    for i in tqdm(range(n_reps), desc='Shuffling repetitions'):
        anova_cells_shuffled, _, _, R_S, pref_num_shuffled, excitatory_or_inhibitory_S = compute_shuffled_anova_neurons(
            Q, C, Hf, alpha_level, n_permutations, show_inner_progress)

        # Get the types of selected neurons directly
        selected_neuron_types = excitatory_or_inhibitory_S
        selected_preferred_numerosities = pref_num_shuffled

        # Count excitatory and inhibitory neurons for each preferred numerosity level
        for q in range(n_numerosities):
            count_exc = np.sum(selected_preferred_numerosities[selected_neuron_types == 'excitatory'] == q)
            count_inh = np.sum(selected_preferred_numerosities[selected_neuron_types == 'inhibitory'] == q)
            count_total = np.sum(selected_preferred_numerosities == q)

            n_neurons_shuffled_excitatory[q, i] = count_exc
            n_neurons_shuffled_inhibitory[q, i] = count_inh
            n_neurons_shuffled_total[q, i] = count_total

    # Calculate the mean and SEM for both neuron types and total
    chance_means_excitatory = np.mean(n_neurons_shuffled_excitatory, axis=1)
    chance_sems_excitatory = np.std(n_neurons_shuffled_excitatory, axis=1) / np.sqrt(n_reps)

    chance_means_inhibitory = np.mean(n_neurons_shuffled_inhibitory, axis=1)
    chance_sems_inhibitory = np.std(n_neurons_shuffled_inhibitory, axis=1) / np.sqrt(n_reps)

    chance_means_total = np.mean(n_neurons_shuffled_total, axis=1)
    chance_sems_total = np.std(n_neurons_shuffled_total, axis=1) / np.sqrt(n_reps)

    chance_means = {
        'excitatory': chance_means_excitatory,
        'inhibitory': chance_means_inhibitory,
        'total': chance_means_total
    }
    
    chance_sems = {
        'excitatory': chance_sems_excitatory,
        'inhibitory': chance_sems_inhibitory,
        'total': chance_sems_total
    }

    for q in range(n_numerosities):
        print(f"Preferred Numerosity {q}:")
        print(f"  Excitatory chance level: {int(round(chance_means_excitatory[q]))} ± {int(round(chance_sems_excitatory[q]))}")
        print(f"  Inhibitory chance level: {int(round(chance_means_inhibitory[q]))} ± {int(round(chance_sems_inhibitory[q]))}")
        print(f"  Total chance level: {int(round(chance_means_total[q]))} ± {int(round(chance_sems_total[q]))}")

    return chance_means, chance_sems

# CREATE ANOVA CELLS TIF VOLUMES TO VISUALIZE 3D CELLS DISTRIBUTION
def save_anova_mask(spots, spot_tag, region_tag, vol_size, resolution):
    # vol_size and resolution ordered as: (Z, Y, X)
    print('Creating tiff with cells in 3D volume...')
    anova_group_tags = ["anova_groups", "anova_num_0", "anova_num_1","anova_num_2","anova_num_3","anova_num_4","anova_num_5"]
    for anova_tag in anova_group_tags:

        r = spots.groups[region_tag]
        a = spots.groups[anova_tag]
        combined_group = group_logical_and(r, a)    
        mask = spots.get_group_mask(combined_group, vol_size) #spots.groups[combined_tag], (Z, Y, X))

        tif.imwrite(f'./spots/masks/mask_from_{spot_tag}_{region_tag}_{anova_tag}.tif',
                                mask.astype(np.uint16), shape=vol_size,
                                metadata={'spacing': resolution[0], 'unit': 'um', 'axes': 'ZYX'},
                                resolution=(1 / resolution[1], 1 / resolution[2]), imagej=True)


def save_anova_spots(spots, anova_df, spot_tag):

    print('Saving spots with neurons info...')

    anova_cell = anova_df["anova_cells"].values
    pref_num = anova_df["pref_num"].values.astype(int)
    #create anova group
    anova_groups = np.zeros(spots.num_spots)
    anova_groups[anova_cell] = 1
    #create group numerosity 0
    anova_num_0 = np.zeros(spots.num_spots)
    anova_num_0[anova_cell[pref_num == 0]] = 1
    #create group numerosity 1
    anova_num_1 = np.zeros(spots.num_spots)
    anova_num_1[anova_cell[pref_num == 1]] = 1
    #create group numerosity 2
    anova_num_2 = np.zeros(spots.num_spots)
    anova_num_2[anova_cell[pref_num == 2]] = 1
    #create group numerosity 3
    anova_num_3 = np.zeros(spots.num_spots)
    anova_num_3[anova_cell[pref_num == 3]] = 1
    #create group numerosity 4
    anova_num_4 = np.zeros(spots.num_spots)
    anova_num_4[anova_cell[pref_num == 4]] = 1
    #create group numerosity 5
    anova_num_5 = np.zeros(spots.num_spots)
    anova_num_5[anova_cell[pref_num == 5]] = 1


    spots.add_groups({"anova_groups":anova_groups,
                    "anova_num_0":anova_num_0,
                    "anova_num_1":anova_num_1, 
                    "anova_num_2":anova_num_2, 
                    "anova_num_3":anova_num_3, 
                    "anova_num_4":anova_num_4, 
                    "anova_num_5":anova_num_5}, 
                    rewrite=True)

    spots.to_json(f"./spots/signals/spots_{spot_tag}.json")



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from statistics import mean, stdev, sqrt
import scipy.stats as stats
import os

def average_tuning_curves(Q, H):
    #Q: array of type of stimulus (trials,)
    #H: matrix of responses (trials, cells)
    Qrange = np.unique(Q) #Qrange: array of unique type of stimuli (n_numerosities,)
    tuning_curves = np.array([H[Q==j,:].mean(axis=0) for j in Qrange]) # (n_num, cells)
    
    return tuning_curves

def preferred_numerosity(Q, H):
    tuning_curves = average_tuning_curves(Q, H)

    #pref_num = np.unique(Q)[np.argmax(tuning_curves, axis=0)]
    # added abs to consider possible inhibitions!!!!
    pref_num = np.unique(Q)[np.argmax(np.abs(tuning_curves), axis=0)]

    # Find if activity is excitatory (positive) or inhibitory (negative)
    max_values = tuning_curves[np.argmax(np.abs(tuning_curves), axis=0), np.arange(tuning_curves.shape[1])]
    excitatory_or_inhibitory = np.where(max_values > 0, 'excitatory', 'inhibitory')
    
    return pref_num, excitatory_or_inhibitory

def get_tuning_matrix(Q, R, pref_num, excitatory_or_inhibitory, n_numerosities):
    # 1. Calculate average tuning curve of each unit
    tuning_curves = average_tuning_curves(Q, R)
    
    # Arrays to store results for excitatory and inhibitory neurons
    tuning_mat_exc = []
    tuning_mat_inh = []
    tuning_err_exc = []
    tuning_err_inh = []

    # 2. Calculate population tuning curves separately for excitatory and inhibitory neurons
    for q in np.arange(n_numerosities):
        # For excitatory neurons
        exc_indices = np.logical_and(pref_num == q, excitatory_or_inhibitory == 'excitatory')
        if np.any(exc_indices):
            tuning_mat_exc.append(np.mean(tuning_curves[:, exc_indices], axis=1))
            tuning_err_exc.append(np.std(tuning_curves[:, exc_indices], axis=1) / np.sqrt(np.sum(exc_indices)))
        else:
            tuning_mat_exc.append(np.zeros(tuning_curves.shape[0]))  # Append zeros if no excitatory neurons found
            tuning_err_exc.append(np.zeros(tuning_curves.shape[0]))

        # For inhibitory neurons
        inh_indices = np.logical_and(pref_num == q, excitatory_or_inhibitory == 'inhibitory')
        if np.any(inh_indices):
            tuning_mat_inh.append(np.mean(tuning_curves[:, inh_indices], axis=1))
            tuning_err_inh.append(np.std(tuning_curves[:, inh_indices], axis=1) / np.sqrt(np.sum(inh_indices)))
        else:
            tuning_mat_inh.append(np.zeros(tuning_curves.shape[0]))  # Append zeros if no inhibitory neurons found
            tuning_err_inh.append(np.zeros(tuning_curves.shape[0]))

    # Convert lists to numpy arrays
    tuning_mat_exc = np.array(tuning_mat_exc)
    tuning_err_exc = np.array(tuning_err_exc)
    tuning_mat_inh = np.array(tuning_mat_inh)
    tuning_err_inh = np.array(tuning_err_inh)

    # 3. Normalize population tuning curves to the 0-1 range for both excitatory and inhibitory neurons
    def normalize_tuning(tuning_mat, tuning_err):
        tmmin = tuning_mat.min(axis=1)[:, None]
        tmmax = tuning_mat.max(axis=1)[:, None]
        tuning_mat_norm = (tuning_mat - tmmin) / (tmmax - tmmin)
        tuning_err_norm = tuning_err / (tmmax - tmmin)
        return tuning_mat_norm, tuning_err_norm

    tuning_mat_exc, tuning_err_exc = normalize_tuning(tuning_mat_exc, tuning_err_exc)
    tuning_mat_inh, tuning_err_inh = normalize_tuning(tuning_mat_inh, tuning_err_inh)

    return tuning_mat_exc, tuning_err_exc, tuning_mat_inh, tuning_err_inh

def plot_bar_with_error(ax, Qrange, hist, chance_means, chance_errors, title, xlabel, ylabel, colors_list):
    """Helper function to plot bar chart with error shading and chance means as horizontal bars."""
    perc = hist / np.sum(hist)
    ax.bar(Qrange, hist, width=0.9, color=colors_list, alpha=0.7)  # Added alpha for transparency

    # Update text positioning and formatting
    for x, y, p in zip(Qrange, hist, perc):
        ax.text(x - 0.4, 0, str(y) + '\n' + str(round(p * 100, 1)) + '%', ha='left', va='bottom', fontsize=9)  # Positioning text inside the canvas

    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    if chance_means is not None and chance_errors is not None:
        for q in Qrange:
            # Draw horizontal line for chance mean
            ax.hlines(y=chance_means[q], xmin=q - 0.4, xmax=q + 0.4, color='black', linewidth=2, label='Chance Mean' if q == 0 else "")
            # Draw error bar as horizontal line above the chance mean
            ax.hlines(y=chance_means[q] - chance_errors[q], xmin=q - 0.4, xmax=q + 0.4, color='black', linestyle='--', linewidth=1)
            ax.hlines(y=chance_means[q] + chance_errors[q], xmin=q - 0.4, xmax=q + 0.4, color='black', linestyle='--', linewidth=1)

    plt.xticks(Qrange)


def plot_selective_cells_histo(pref_num, n_numerosities, colors_list, excitatory_or_inhibitory=None, chance_means=None, chance_errors=None, file_name=None):
    """Main function to plot selective cells histogram with simplified save path."""
    Qrange = np.arange(n_numerosities)

    if excitatory_or_inhibitory is None:
        # Case for all neurons
        hist = [np.sum(pref_num == q) for q in Qrange]
        fig, ax = plt.subplots(figsize=(4, 4))
        plot_bar_with_error(ax, Qrange, hist, chance_means, chance_errors, "All Neurons", 'Preferred Numerosity', 'Number of cells', colors_list)
        plt.xticks(Qrange)
        plt.legend()
    else:
        # Case with excitatory and inhibitory neurons
        fig, axes = plt.subplots(1, 3, figsize=(12, 4))

        # Excitatory neurons
        excitatory_indices = excitatory_or_inhibitory == 'excitatory'
        hist_exc = [np.sum(pref_num[excitatory_indices] == q) for q in Qrange]
        plot_bar_with_error(axes[0], Qrange, hist_exc, 
                            chance_means['excitatory'], 
                            chance_errors['excitatory'], 
                            'Excitatory Neurons', 'Preferred Numerosity', 'Number of cells', colors_list)

        # Inhibitory neurons
        inhibitory_indices = excitatory_or_inhibitory == 'inhibitory'
        hist_inh = [np.sum(pref_num[inhibitory_indices] == q) for q in Qrange]
        plot_bar_with_error(axes[1], Qrange, hist_inh, 
                            chance_means['inhibitory'], 
                            chance_errors['inhibitory'], 
                            'Inhibitory Neurons', 'Preferred Numerosity', 'Number of cells', colors_list)

        # All neurons
        hist = [np.sum(pref_num == q) for q in Qrange]
        plot_bar_with_error(axes[2], Qrange, hist, 
                            chance_means['total'], 
                            chance_errors['total'], 
                            'All Neurons', 'Preferred Numerosity', 'Number of cells', colors_list)

        plt.tight_layout()

    # Save figure
    if file_name:
        try:
            if file_name.endswith('.svg'):
                plt.savefig(file_name)
            elif file_name.endswith('.png'):
                plt.savefig(file_name, dpi=900)
            else:
                print("Error: file_name should end with either '.svg' or '.png'")
                return
            #print(f"File saved successfully at: {os.path.abspath(file_name)}")
        except Exception as e:
            print(f"Error while saving file: {e}")

    plt.show()


def plot_tuning_curves(tuning_mat_exc, tuning_err_exc,  colors=None, tuning_mat_inh=None, tuning_err_inh=None, excitatory_or_inhibitory=None):
    # Number of types of stimuli (should be the same for both matrices)
    n_stimuli = tuning_mat_exc.shape[0]  # This should match tuning_mat_inh if provided

    # Check if the color list is provided
    if colors is None:
        colors = ['red', 'blue', 'green', 'orange', 'purple', 'cyan']  # Default colors

    # Create a single plot for both categories if both are provided
    if excitatory_or_inhibitory is not None:
        # Create a figure with subplots
        fig, axs = plt.subplots(1, 2, figsize=(12, 5))

        # Plot for excitatory neurons
        for i in range(n_stimuli):
            axs[0].plot(np.arange(n_stimuli), tuning_mat_exc[i], color=colors[i], label=f'{i}')
            axs[0].fill_between(np.arange(n_stimuli), 
                                tuning_mat_exc[i] - tuning_err_exc[i], 
                                tuning_mat_exc[i] + tuning_err_exc[i], 
                                color=colors[i], alpha=0.2)  # Error bars
        axs[0].set_title('Tuning Curves - Excitatory Neurons')
        axs[0].set_xlabel('Numerosity')
        axs[0].set_ylabel('Response')

        # Plot for inhibitory neurons
        if tuning_mat_inh is not None and tuning_err_inh is not None:
            for i in range(n_stimuli):
                axs[1].plot(np.arange(n_stimuli), tuning_mat_inh[i], color=colors[i], label=f'{i}')
                axs[1].fill_between(np.arange(n_stimuli), 
                                    tuning_mat_inh[i] - tuning_err_inh[i], 
                                    tuning_mat_inh[i] + tuning_err_inh[i], 
                                    color=colors[i], alpha=0.2)  # Error bars
            axs[1].set_title('Tuning Curves - Inhibitory Neurons')
            axs[1].set_xlabel('Numerosity')
            axs[1].set_ylabel('Response')

        # Adjust layout and show legend
        plt.tight_layout()
        plt.show()

    else:
        # If no classification provided, plot only excitatory neurons
        plt.figure(figsize=(10, 5))
        for i in range(n_stimuli):
            plt.plot(np.arange(n_stimuli), tuning_mat_exc[i], color=colors[i], label=f'{i}')
            plt.fill_between(np.arange(n_stimuli), 
                             tuning_mat_exc[i] - tuning_err_exc[i], 
                             tuning_mat_exc[i] + tuning_err_exc[i], 
                             color=colors[i], alpha=0.2)  # Error bars
        plt.title('Tuning Curves - Excitatory Neurons')
        plt.xlabel('Numerosity')
        plt.ylabel('Response')
        plt.legend(title='Stimuli')
        plt.show()
    
def populate_tuning_dicts(tuning_mat_exc, tuning_mat_inh, n_numerosities):
    distRange_abs_0 = np.arange(-(n_numerosities-1), n_numerosities).tolist()
    distRange_abs_1 = np.arange(n_numerosities).tolist()

    dist_tuning_dict_abs_0_exc = {str(i): [] for i in distRange_abs_0}
    dist_tuning_dict_abs_1_exc = {str(i): [] for i in distRange_abs_1}
    
    dist_tuning_dict_abs_0_inh = {str(i): [] for i in distRange_abs_0} if tuning_mat_inh is not None else None
    dist_tuning_dict_abs_1_inh = {str(i): [] for i in distRange_abs_1} if tuning_mat_inh is not None else None

    for pref_n in np.arange(n_numerosities):
        for n in np.arange(n_numerosities):
            dist_tuning_dict_abs_0_exc[str(n - pref_n)].append(tuning_mat_exc[pref_n][n])
            dist_tuning_dict_abs_1_exc[str(abs(n - pref_n))].append(tuning_mat_exc[pref_n][n])
            if tuning_mat_inh is not None:
                dist_tuning_dict_abs_0_inh[str(n - pref_n)].append(tuning_mat_inh[pref_n][n])
                dist_tuning_dict_abs_1_inh[str(abs(n - pref_n))].append(tuning_mat_inh[pref_n][n])

    return (distRange_abs_0, distRange_abs_1, dist_tuning_dict_abs_0_exc, dist_tuning_dict_abs_1_exc, dist_tuning_dict_abs_0_inh, dist_tuning_dict_abs_1_inh)

def calculate_avg_and_err(dist_tuning_dict):
    avg_tuning = [np.mean(dist_tuning_dict[key]) if dist_tuning_dict[key] else 0 for key in dist_tuning_dict.keys()]
    err_tuning = [np.nanstd(dist_tuning_dict[key]) / sqrt(len(dist_tuning_dict[key])) if len(dist_tuning_dict[key]) > 1 else 0 for key in dist_tuning_dict.keys()]
    return avg_tuning, err_tuning

def plot_avg_tunings(distRange_abs_0, distRange_abs_1, dist_avg_tuning_abs_0_exc, dist_err_tuning_abs_0_exc, dist_avg_tuning_abs_1_exc, dist_err_tuning_abs_1_exc, dist_avg_tuning_abs_0_inh, dist_err_tuning_abs_0_inh, dist_avg_tuning_abs_1_inh, dist_err_tuning_abs_1_inh, save_file=None):
    num_plots = 2 if dist_avg_tuning_abs_0_inh is None else 4
    fig, axs = plt.subplots(2, 2, figsize=(10, 8)) if dist_avg_tuning_abs_0_inh is not None else plt.subplots(1, 2, figsize=(10, 5))
    axs = axs.flatten() if dist_avg_tuning_abs_0_inh is not None else axs
    
    # Plot excitatory
    axs[0].errorbar(distRange_abs_0, dist_avg_tuning_abs_0_exc, yerr=dist_err_tuning_abs_0_exc, color='black')
    axs[0].set_title('Excitatory Neurons: Numerical Distance')
    
    axs[1].errorbar(distRange_abs_1, dist_avg_tuning_abs_1_exc, yerr=dist_err_tuning_abs_1_exc, color='black')
    axs[1].set_title('Excitatory Neurons: Absolute Distance')
    
    # Plot inhibitory if present
    if dist_avg_tuning_abs_0_inh is not None:
        axs[2].errorbar(distRange_abs_0, dist_avg_tuning_abs_0_inh, yerr=dist_err_tuning_abs_0_inh, color='black')
        axs[2].set_title('Inhibitory Neurons: Numerical Distance')

        axs[3].errorbar(distRange_abs_1, dist_avg_tuning_abs_1_inh, yerr=dist_err_tuning_abs_1_inh, color='black')
        axs[3].set_title('Inhibitory Neurons: Absolute Distance')
    
    plt.tight_layout()
    if save_file:
        plt.savefig(f'{save_file}.png')
    plt.show()

def t_test_to_dataframe(dist_tuning_dict_abs_0, dist_tuning_dict_abs_1, n_numerosities, file_name=None):
    distance_comparisons = [(i, i + 1) for i in range(-n_numerosities + 1, n_numerosities - 1)]
    
    data = {
        'Distance Pair': [],
        't-stat (d0)': [],
        'p-value (d0)': [],
        'df (d0)': [],
        't-stat (d1)': [],
        'p-value (d1)': [],
        'df (d1)': []
    }
    
    for d1, d2 in distance_comparisons:
        # For numerical distance (d0)
        if str(d1) in dist_tuning_dict_abs_0 and str(d2) in dist_tuning_dict_abs_0:
            t_stat, p_value = stats.ttest_ind(a=dist_tuning_dict_abs_0[str(d1)], b=dist_tuning_dict_abs_0[str(d2)], equal_var=False)
            df = len(dist_tuning_dict_abs_0[str(d1)]) + len(dist_tuning_dict_abs_0[str(d2)]) - 2
            data['Distance Pair'].append(f'{d1} vs {d2}')
            data['t-stat (d0)'].append(t_stat)
            data['p-value (d0)'].append(p_value)
            data['df (d0)'].append(df)
        else:
            data['Distance Pair'].append(f'{d1} vs {d2}')
            data['t-stat (d0)'].append(None)
            data['p-value (d0)'].append(None)
            data['df (d0)'].append(None)

        # For absolute distance (d1), only calculate for positive comparisons
        if d1 >= 0:  # Only consider non-negative comparisons for d1
            if str(abs(d1)) in dist_tuning_dict_abs_1 and str(abs(d2)) in dist_tuning_dict_abs_1:
                t_stat, p_value = stats.ttest_ind(a=dist_tuning_dict_abs_1[str(abs(d1))], b=dist_tuning_dict_abs_1[str(abs(d2))], equal_var=False)
                df = len(dist_tuning_dict_abs_1[str(abs(d1))]) + len(dist_tuning_dict_abs_1[str(abs(d2))]) - 2
                data['t-stat (d1)'].append(t_stat)
                data['p-value (d1)'].append(p_value)
                data['df (d1)'].append(df)
            else:
                data['t-stat (d1)'].append(None)
                data['p-value (d1)'].append(None)
                data['df (d1)'].append(None)
        else:
            # Fill with None for negative comparisons in d1
            data['t-stat (d1)'].append(None)
            data['p-value (d1)'].append(None)
            data['df (d1)'].append(None)
    
    df = pd.DataFrame(data)
    
    # Save to Excel or CSV if filename is provided
    if file_name:
        try:
            if file_name.endswith('.xlsx'):
                df.to_excel(file_name, index=False)
            elif file_name.endswith('.csv'):
                df.to_csv(file_name, index=False)
            else:
                print("Error: file_name should end with either '.xlsx' or '.csv'")
                return
            #print(f"File saved successfully at: {os.path.abspath(file_name)}")
        except Exception as e:
            print(f"Error while saving file: {e}")

    return df


def plot_abs_dist_tunings(tuning_mat_exc, n_numerosities, tuning_mat_inh=None, save_file=None, print_stats=True, plot_figures=True):
    # Step 1: Populate tuning dictionaries
    distRange_abs_0, distRange_abs_1, dist_tuning_dict_abs_0_exc, dist_tuning_dict_abs_1_exc, dist_tuning_dict_abs_0_inh, dist_tuning_dict_abs_1_inh = populate_tuning_dicts(tuning_mat_exc, tuning_mat_inh, n_numerosities)
    
    # Step 2: Calculate averages and errors for excitatory neurons
    dist_avg_tuning_abs_0_exc, dist_err_tuning_abs_0_exc = calculate_avg_and_err(dist_tuning_dict_abs_0_exc)
    dist_avg_tuning_abs_1_exc, dist_err_tuning_abs_1_exc = calculate_avg_and_err(dist_tuning_dict_abs_1_exc)
    
    # Step 3: Calculate averages and errors for inhibitory neurons if provided
    if tuning_mat_inh is not None:
        dist_avg_tuning_abs_0_inh, dist_err_tuning_abs_0_inh = calculate_avg_and_err(dist_tuning_dict_abs_0_inh)
        dist_avg_tuning_abs_1_inh, dist_err_tuning_abs_1_inh = calculate_avg_and_err(dist_tuning_dict_abs_1_inh)
    else:
        dist_avg_tuning_abs_0_inh = dist_avg_tuning_abs_1_inh = None
        dist_err_tuning_abs_0_inh = dist_err_tuning_abs_1_inh = None
    
    # Step 4: Plot figures (if plot_figures is True)
    if plot_figures:
        plot_avg_tunings(distRange_abs_0, distRange_abs_1, dist_avg_tuning_abs_0_exc, dist_err_tuning_abs_0_exc, dist_avg_tuning_abs_1_exc, dist_err_tuning_abs_1_exc, dist_avg_tuning_abs_0_inh, dist_err_tuning_abs_0_inh, dist_avg_tuning_abs_1_inh, dist_err_tuning_abs_1_inh, save_file)
    
    # Step 5: Print stats if enabled
    df_exc = t_test_to_dataframe(dist_tuning_dict_abs_0_exc, dist_tuning_dict_abs_1_exc, n_numerosities,'./anova_results/excitatory_average_tunings.csv')
    if print_stats:
        print("Excitatory Neurons t-tests:")
        print(df_exc)
    if tuning_mat_inh is not None:
        df_inh = t_test_to_dataframe(dist_tuning_dict_abs_0_inh, dist_tuning_dict_abs_1_inh, n_numerosities, './anova_results/inhibitory_average_tunings.csv')
        if print_stats:
            print("\nInhibitory Neurons t-tests:")
            print(df_inh)
    
    # Step 6: Return results
    return {
        'exc_avg_tuning_abs_0': dist_avg_tuning_abs_0_exc,
        'exc_err_tuning_abs_0': dist_err_tuning_abs_0_exc,
        'exc_avg_tuning_abs_1': dist_avg_tuning_abs_1_exc,
        'exc_err_tuning_abs_1': dist_err_tuning_abs_1_exc,
        'inh_avg_tuning_abs_0': dist_avg_tuning_abs_0_inh,
        'inh_err_tuning_abs_0': dist_err_tuning_abs_0_inh,
        'inh_avg_tuning_abs_1': dist_avg_tuning_abs_1_inh,
        'inh_err_tuning_abs_1': dist_err_tuning_abs_1_inh,
        'distRange_abs_0': distRange_abs_0,
        'distRange_abs_1': distRange_abs_1
    }

def replot_tuning_curves(output_real, output_shuffled):
    # Extract data from the real output dictionary
    exc_avg_tuning_abs_0_real = output_real['exc_avg_tuning_abs_0']
    exc_err_tuning_abs_0_real = output_real['exc_err_tuning_abs_0']
    exc_avg_tuning_abs_1_real = output_real['exc_avg_tuning_abs_1']
    exc_err_tuning_abs_1_real = output_real['exc_err_tuning_abs_1']
    inh_avg_tuning_abs_0_real = output_real['inh_avg_tuning_abs_0']
    inh_err_tuning_abs_0_real = output_real['inh_err_tuning_abs_0']
    inh_avg_tuning_abs_1_real = output_real['inh_avg_tuning_abs_1']
    inh_err_tuning_abs_1_real = output_real['inh_err_tuning_abs_1']
    distRange_abs_0_real = output_real['distRange_abs_0']
    distRange_abs_1_real = output_real['distRange_abs_1']

    # Extract data from the shuffled output dictionary
    exc_avg_tuning_abs_0_shuffled = output_shuffled['exc_avg_tuning_abs_0']
    exc_err_tuning_abs_0_shuffled = output_shuffled['exc_err_tuning_abs_0']
    exc_avg_tuning_abs_1_shuffled = output_shuffled['exc_avg_tuning_abs_1']
    exc_err_tuning_abs_1_shuffled = output_shuffled['exc_err_tuning_abs_1']
    inh_avg_tuning_abs_0_shuffled = output_shuffled['inh_avg_tuning_abs_0']
    inh_err_tuning_abs_0_shuffled = output_shuffled['inh_err_tuning_abs_0']
    inh_avg_tuning_abs_1_shuffled = output_shuffled['inh_avg_tuning_abs_1']
    inh_err_tuning_abs_1_shuffled = output_shuffled['inh_err_tuning_abs_1']
    distRange_abs_0_shuffled = output_shuffled['distRange_abs_0']
    distRange_abs_1_shuffled = output_shuffled['distRange_abs_1']

    # Set up the figure with a number of subplots
    num_plots = 2  # Initialize number of plots for excitatory neurons
    if inh_avg_tuning_abs_0_real is not None:  # Check if inhibitory data is available
        num_plots += 2  # Update number of plots to include inhibitory neurons
        fig, axs = plt.subplots(2, 2, figsize=(10, 8))  # 2x2 layout
        axs = axs.flatten()  # Flatten for easier indexing
    else:
        fig, axs = plt.subplots(1, 2, figsize=(10, 5))  # 1x2 layout

    # Plot for excitatory neurons - numerical distance = 0
    axs[0].errorbar(distRange_abs_0_real, exc_avg_tuning_abs_0_real, 
                    yerr=exc_err_tuning_abs_0_real, 
                    color='blue', label='Real Data', capsize=3)
    axs[0].errorbar(distRange_abs_0_shuffled, exc_avg_tuning_abs_0_shuffled, 
                    yerr=exc_err_tuning_abs_0_shuffled, 
                    color='cyan', label='Shuffled Data', capsize=3)
    axs[0].set_xticks(distRange_abs_0_real)
    axs[0].set_xlabel('Numerical Distance')
    axs[0].set_ylabel('Normalized Neural Activity')
    axs[0].set_title('Numerical Distance Tuning Curve (Excitatory)')
    axs[0].legend()

    # Plot for excitatory neurons - absolute distance = 1
    axs[1].errorbar(distRange_abs_1_real, exc_avg_tuning_abs_1_real, 
                    yerr=exc_err_tuning_abs_1_real, 
                    color='blue', label='Real Data', capsize=3)
    axs[1].errorbar(distRange_abs_1_shuffled, exc_avg_tuning_abs_1_shuffled, 
                    yerr=exc_err_tuning_abs_1_shuffled, 
                    color='cyan', label='Shuffled Data', capsize=3)
    axs[1].set_xticks(distRange_abs_1_real)
    axs[1].set_xlabel('Absolute Numerical Distance')
    axs[1].set_ylabel('Normalized Neural Activity')
    axs[1].set_title('Absolute Numerical Distance Tuning Curve (Excitatory)')
    axs[1].legend()

    # If inhibitory neurons are provided, create their plots
    if inh_avg_tuning_abs_0_real is not None:
        # Plot for inhibitory neurons - numerical distance = 0
        axs[2].errorbar(distRange_abs_0_real, inh_avg_tuning_abs_0_real, 
                        yerr=inh_err_tuning_abs_0_real, 
                        color='red', label='Real Data', capsize=3)
        axs[2].errorbar(distRange_abs_0_shuffled, inh_avg_tuning_abs_0_shuffled, 
                        yerr=inh_err_tuning_abs_0_shuffled, 
                        color='orange', label='Shuffled Data', capsize=3)
        axs[2].set_xticks(distRange_abs_0_real)
        axs[2].set_xlabel('Numerical Distance')
        axs[2].set_ylabel('Normalized Neural Activity')
        axs[2].set_title('Numerical Distance Tuning Curve (Inhibitory)')
        axs[2].legend()

        # Plot for inhibitory neurons - absolute distance = 1
        axs[3].errorbar(distRange_abs_1_real, inh_avg_tuning_abs_1_real, 
                        yerr=inh_err_tuning_abs_1_real, 
                        color='red', label='Real Data', capsize=3)
        axs[3].errorbar(distRange_abs_1_shuffled, inh_avg_tuning_abs_1_shuffled, 
                        yerr=inh_err_tuning_abs_1_shuffled, 
                        color='orange', label='Shuffled Data', capsize=3)
        axs[3].set_xticks(distRange_abs_1_real)
        axs[3].set_xlabel('Absolute Numerical Distance')
        axs[3].set_ylabel('Normalized Neural Activity')
        axs[3].set_title('Absolute Numerical Distance Tuning Curve (Inhibitory)')
        axs[3].legend()

    plt.tight_layout()
    plt.show()


def plot_tunings(Q, R, Q_S, R_S, brain_region_tag, chance_mean, chance_error, n_numerosities, colors_list, save_file=None, print_stats=True, plot_figures=True):

    # Create new folder to save ANOVA graphs
    os.makedirs('./anova_figures', exist_ok=True)
    save_path = './anova_figures'

    # Real data ######################################################################
    print('\033[1m\nREAL DATA\033[0m\n')

    pref_num, excitatory_or_inhibitory = preferred_numerosity(Q, R)
 
    tuning_mat_exc, tuning_err_exc, tuning_mat_inh, tuning_err_inh = get_tuning_matrix(
        Q, R, pref_num, excitatory_or_inhibitory, n_numerosities
    )
    
    plot_tuning_curves(
        tuning_mat_exc, tuning_err_exc, colors_list, 
        tuning_mat_inh, tuning_err_inh, excitatory_or_inhibitory
    )
    output = plot_abs_dist_tunings(tuning_mat_exc, n_numerosities, tuning_mat_inh, save_file, print_stats, plot_figures)

    
    # Shuffled data #####################################################################
    print('\033[1m\nSHUFFLED DATA\033[0m\n')

    pref_num_shuffled, excitatory_or_inhibitory_shuffled = preferred_numerosity(Q_S, R_S)
    
    tuning_mat_exc_shuffled, tuning_err_exc_shuffled, tuning_mat_inh_shuffled, tuning_err_inh_shuffled = get_tuning_matrix(
        Q_S, R_S, pref_num_shuffled, excitatory_or_inhibitory_shuffled, n_numerosities
    )

    plot_tuning_curves(
        tuning_mat_exc_shuffled, tuning_err_exc_shuffled, colors_list, 
        tuning_mat_inh_shuffled, tuning_err_inh_shuffled, excitatory_or_inhibitory_shuffled
    )
    output_shuffled = plot_abs_dist_tunings(
        tuning_mat_exc_shuffled, n_numerosities, tuning_mat_inh_shuffled, 
        save_file=None, print_stats=False, plot_figures=False
    )

    # Plot real and shuffled tuning curves together
    print('\nOVERALL TUNINGS BASED ON NUMERICAL DISTANCE')
    replot_tuning_curves(output, output_shuffled)